from collections.abc import Sequence
from datetime import datetime
from typing import Generic, TypeVar

from pydantic import BaseModel, ConfigDict, Field

from brickworks.core.utils.timeutils import convert_datetime_to_iso_8601_with_z_suffix


class BaseSchema(BaseModel):
    """
    Base schema for all schemas in the project.
    It converts datetime objects to ISO 8601 format with Z suffix.
    Make sure that your datetime objects are timezone unaware!
    """

    model_config = ConfigDict(
        json_encoders={datetime: convert_datetime_to_iso_8601_with_z_suffix}, from_attributes=True
    )


T = TypeVar("T")


class PaginatedResponse(BaseModel, Generic[T]):
    items: Sequence[T] = Field(..., description="List of items on this page.")
    total: int = Field(..., description="Total number of items available.")
    page: int = Field(..., description="Current page number (1-based).")
    page_size: int = Field(..., description="Number of items per page.")



from typing import Any

from fastapi import APIRouter

from brickworks.core.exceptions import NotFoundException
from brickworks.core.models.base_dbmodel import BaseDBModel
from brickworks.core.models.base_view import BaseView
from brickworks.core.schemas.base_schema import PaginatedResponse


class WithListRoute:
    """
    Mixin to add a paginated list GET route to any class with a get_list_with_policies classmethod.
    """

    __routing_path__: str  # The path for the GET route, must be defined in the subclass.

    def __init_subclass__(cls, **kwargs: Any) -> None:  # noqa: ANN401
        super().__init_subclass__(**kwargs)
        if not issubclass(cls, BaseView) and not issubclass(cls, BaseDBModel):
            raise TypeError(f"{cls.__name__} must be a subclass of BaseView or BaseDBModel to use WithListRouteMixin.")
        if not hasattr(cls, "__routing_path__"):
            raise ValueError(f"{cls.__name__} must define a __routing_path__ attribute to use WithListRouteMixin.")
        if not cls.__routing_path__:
            raise ValueError(f"{cls.__name__} must define a __routing_path__ attribute to use WithListRouteMixin.")

    @classmethod
    def get_list_router(cls) -> APIRouter:
        if not issubclass(cls, BaseView) and not issubclass(cls, BaseDBModel):
            raise TypeError(f"{cls.__name__} must be a subclass of BaseView or BaseDBModel to use WithListRouteMixin.")

        router = APIRouter()

        async def _get_all(page: int = 1, page_size: int = 500) -> PaginatedResponse[BaseDBModel | BaseView]:
            items, total = await cls.get_paginated_list_with_policies(_per_page=page_size, _page=page)
            return PaginatedResponse(items=items, total=total, page=page, page_size=page_size)

        router.add_api_route(
            cls.__routing_path__,
            _get_all,
            response_model=PaginatedResponse[cls],  # type: ignore
            summary=f"Get all {cls.__name__} objects (paginated)",
            description=f"Get all {cls.__name__} objects with pagination.",
        )
        return router


class WithGetRoute:
    """
    Mixin to add a GET-by-key route to any class with a get_one_or_none_with_policies classmethod.
    """

    __routing_path__: str  # The path for the GET route, must be defined in the subclass.
    __routing_get_key__: str  # Key to use for the GET route, must be defined in the subclass.

    def __init_subclass__(cls, **kwargs: Any) -> None:  # noqa: ANN401
        super().__init_subclass__(**kwargs)
        if not issubclass(cls, BaseView) and not issubclass(cls, BaseDBModel):
            raise TypeError(f"{cls.__name__} must be a subclass of BaseView or BaseDBModel to use WithGetRouteMixin.")
        if not hasattr(cls, "__routing_path__"):
            raise ValueError(f"{cls.__name__} must define a __routing_path__ attribute to use WithGetRouteMixin.")
        if not cls.__routing_path__:
            raise ValueError(f"{cls.__name__} must define a __routing_path__ attribute to use WithGetRouteMixin.")
        if not hasattr(cls, "__routing_get_key__") or not cls.__routing_get_key__:
            raise ValueError(f"{cls.__name__} must define a __routing_get_key__ attribute to use WithGetRouteMixin.")

    @classmethod
    def get_get_router(cls) -> APIRouter:
        if not issubclass(cls, BaseView) and not issubclass(cls, BaseDBModel):
            raise TypeError(f"{cls.__name__} must be a subclass of BaseView or BaseDBModel to use WithGetRouteMixin.")

        router = APIRouter()
        routing_get_key = cls.__routing_get_key__

        async def _get_by_key(key: str) -> BaseView | BaseDBModel | None:
            result = await cls.get_one_or_none_with_policies(_filter_clause=None, **{routing_get_key: key})
            if not result:
                raise NotFoundException(f"{cls.__name__} with {routing_get_key} '{key}' not found.")
            return result

        router.add_api_route(
            f"{cls.__routing_path__}/{{key}}",
            _get_by_key,
            response_model=cls | None,
            summary=f"Get a {cls.__name__} object by {cls.__routing_get_key__}",
            description=f"Get a {cls.__name__} object by {cls.__routing_get_key__}.",
        )
        return router



from httpx import AsyncClient
from sqlalchemy import func, select
from sqlalchemy.orm import aliased

from brickworks.core.acl.policies import AllowPublicAccessPolicy
from brickworks.core.models.base_view import BaseView
from brickworks.core.models.mixins import WithGetRoute, WithListRoute
from brickworks.core.models.role_model import RoleModel, user_role_table
from brickworks.core.models.user_model import UserModel
from tests.core.utils import create_test_user

user_alias = aliased(UserModel)
user_role_table_alias = aliased(user_role_table)


class RolesPerUserView(BaseView, WithGetRoute, WithListRoute):
    # define the fields of your query result
    # the field names need to match the (labeled) column names returned by the select statement
    user_name: str
    role_count: int

    __routing_path__ = "/test/roles_per_user"
    __routing_get_key__ = "user_name"
    __policies__ = [AllowPublicAccessPolicy()]  # policies to apply to this view, can be empty
    __policy_model_class__ = UserModel  # the model class to apply policies to
    # define the select statement
    __select__ = (
        select(
            user_alias.name.label("user_name"),
            func.count(user_role_table_alias.c.role_uuid).label("role_count"),
        )
        .select_from(user_alias)
        .outerjoin(user_role_table_alias, user_alias.uuid == user_role_table_alias.c.user_uuid)
        .group_by(user_alias.name)
    )


class TestUserModel(UserModel, WithGetRoute, WithListRoute):
    """
    Test UserModel with WithGetRouteMixin for testing purposes.
    """

    __routing_path__ = "/test/user"
    __routing_get_key__ = "name"
    __policies__ = [AllowPublicAccessPolicy()]
    __policy_model_class__ = UserModel


async def test_with_get_route_mixin(client: AsyncClient) -> None:
    """
    Test the WithGetRouteMixin functionality, by using the RolesPerUserView.
    """
    # Create a test user with role
    user = await create_test_user("Alice", "Smith")
    role = await RoleModel(role_name="test_role").persist()
    await user.add_role(role)

    # fetch the roles per user (paginated response)
    response = await client.get("/api/test/roles_per_user?page=1&page_size=300")
    assert response.status_code == 200
    data = response.json()
    assert isinstance(data, dict)
    assert "items" in data and "total" in data and "page" in data and "page_size" in data
    assert isinstance(data["items"], list)
    assert data["total"] >= 1
    assert data["page"] == 1
    assert data["page_size"] == 300

    # check if the user is in the response and has the correct role count
    user_data = next((item for item in data["items"] if item["user_name"] == "Alice Smith"), None)
    assert user_data is not None
    assert user_data["role_count"] == 1

    # fetch the user by name (single object response)
    response = await client.get("/api/test/roles_per_user/Alice Smith")
    assert response.status_code == 200
    data = response.json()
    assert isinstance(data, dict)
    assert data["user_name"] == "Alice Smith"
    assert data["role_count"] == 1


async def test_with_get_route_mixin_model(client: AsyncClient) -> None:
    """
    Test the WithGetRouteMixin functionality with a model class (TestUserModel).
    """
    # Create a test user
    await create_test_user("Bob", "Johnson")

    # fetch the users (paginated response)
    response = await client.get("/api/test/user?page=1&page_size=100")
    assert response.status_code == 200
    data = response.json()
    assert isinstance(data, dict)
    assert "items" in data and "total" in data and "page" in data and "page_size" in data
    assert isinstance(data["items"], list)
    assert data["total"] >= 1
    assert data["page"] == 1
    assert data["page_size"] == 100

    # check if the user is in the response
    user_data = next((item for item in data["items"] if item["name"] == "Bob Johnson"), None)
    assert user_data is not None
    assert user_data["family_name"] == "Johnson"

    # fetch the user by name (single object response)
    response = await client.get("/api/test/user/Bob Johnson")
    assert response.status_code == 200
    data = response.json()
    assert isinstance(data, dict)
    assert data["name"] == "Bob Johnson"
    assert data["family_name"] == "Johnson"



import logging

import uvicorn
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware

from brickworks.core.auth.csrf import CSRFMiddleware
from brickworks.core.auth.executioncontext import ExecutionContextMiddleware
from brickworks.core.auth.session import SessionMiddleware
from brickworks.core.db import DBSessionMiddleware
from brickworks.core.models.mixins import WithGetRoute, WithListRoute
from brickworks.core.module_loader import get_models_by_fqpn, get_views, load_modules
from brickworks.core.utils.importer import import_object_from_path

logging.basicConfig(level=logging.INFO)


def create_app(for_testing: bool = False) -> FastAPI:
    app_base = FastAPI()
    app_api = FastAPI()

    # Middleware is executed in reverse order, so we first add the middleware that should be executed last
    # Middleware loaded from bricks should be executed last, so we add it first
    # (to ensure it has access to sessions and the database)
    for brick in load_modules():
        _add_routers(app_api, brick.routers)
        _add_middlewares(app_api, brick.middlewares)

    _add_auto_routes(app_api)

    if not for_testing:
        # if we run the app with testclient we will create database sessions ourselves, so we can roll back
        app_api.add_middleware(DBSessionMiddleware)

    app_api.add_middleware(CSRFMiddleware)  # needs to be before session middleware
    app_api.add_middleware(
        SessionMiddleware,
        cookie_name="session",
        max_age=14 * 24 * 60 * 60,
        same_site="lax",
        https_only=False,
        domain=None,
    )
    app_api.add_middleware(ExecutionContextMiddleware)
    app_api.add_middleware(
        CORSMiddleware,
        allow_origins=["http://localhost:3000", "http://localhost:8000"],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )
    # app_base.add_event_handler("startup", start_services)

    app_base.mount("/backend/api", app_api)
    return app_base


def _add_routers(app: FastAPI, routers: list[str]) -> None:
    for router_path in routers:
        router = import_object_from_path(router_path)
        app.include_router(router)


def _add_auto_routes(app: FastAPI) -> None:
    views = get_views()
    views_with_get_routes = [view for view in views if issubclass(view, WithGetRoute)]
    for view_with_get in views_with_get_routes:
        router = view_with_get.get_get_router()
        app.include_router(router)

    views_with_list_routes = [view for view in views if issubclass(view, WithListRoute)]
    for view_with_list in views_with_list_routes:
        router = view_with_list.get_list_router()
        app.include_router(router)

    models = get_models_by_fqpn()
    models_with_get_routes = [model for model in models.values() if issubclass(model, WithGetRoute)]
    for model_with_get in models_with_get_routes:
        router = model_with_get.get_get_router()
        app.include_router(router)

    models_with_list_routes = [model for model in models.values() if issubclass(model, WithListRoute)]
    for model_with_list in models_with_list_routes:
        router = model_with_list.get_list_router()
        app.include_router(router)


def _add_middlewares(app: FastAPI, middlewares: list[str]) -> None:
    for middleware_path in middlewares:
        middleware = import_object_from_path(middleware_path)
        app.add_middleware(middleware)


if __name__ == "__main__":
    app = create_app()
    uvicorn.run(app, host="0.0.0.0", port=8000)  # nosec
